import os
import random
import string
import hashlib

from cryptography.fernet import Fernet
from cryptography.hazmat.primitives.asymmetric import rsa, padding
from cryptography.hazmat.primitives import hashes, serialization

from django.conf import settings


class Encryption:

    
    def random_str(n: int = 10) -> str:
        return ''.join(random.choice(string.ascii_uppercase) for _ in range(n))

    
    def random_code(n: int = 10) -> str:
        return ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(n))

    
    def random_hash() -> str:
        return hashlib.sha256(os.urandom(1024)).hexdigest()

    
    def random_int(n: int) -> str:
        ri = random.randint(0, 10 ** n - 1)
        return str(ri)

    
    def aes_key() -> str:
        key = Fernet.generate_key()
        return key.decode()

    
    def aes_encrypt(key: str, plain_text: str) -> str:
        cipher_suite = Fernet(key.encode())
        cipher_text = cipher_suite.encrypt(plain_text.encode())
        return cipher_text.decode()

    
    def aes_decrypt(key: str, cipher_text: str) -> str:
        cipher_suite = Fernet(key.encode())
        plain_text = cipher_suite.decrypt(cipher_text.encode())
        return plain_text.decode()

    def load_private_pem(self) -> str:
        private_pem = None
        with open(os.path.join(settings.BASE_DIR, 'nkunyim_rsa/private_key.pem'), mode='rb') as private_file:
            private_pem = private_file.read()

        return private_pem
    
    
    def load_public_pem(self, name:str) -> str:
        file_name = name.lower()
        public_pem = None
        with open(os.path.join(settings.BASE_DIR, f"nkunyim_rsa/{file_name}_key.pem"), mode='rb') as public_file:
            public_pem = public_file.read()

        return public_pem
    
    
    def rsa_private_key(self) -> rsa.RSAPrivateKey:
        private_key = serialization.load_pem_private_key(
            self.load_private_pem(),
            password=str.encode(settings.RSA_PRIVATE_PHRASE)
        )
        return private_key

    
    def rsa_public_key(self, name:str = None) -> rsa.RSAPublicKey:
        if name:
            public_pem = self.load_public_pem(name=name)
            public_key = serialization.load_pem_public_key(public_pem)
        else:
            private_key = self.rsa_private_key()
            public_key = private_key.public_key()
            
        return public_key
    

    def rsa_encrypt(self, plain_text: str, name:str = None) -> str:
        
        public_key = self.rsa_public_key(name=name)
        
        cipher_text = public_key.encrypt(
            plain_text.encode(),
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None
            )
        )

        return cipher_text


    def rsa_decrypt(self, cipher_text: bytes) -> str:
        private_key = self.rsa_private_key()
        plain_text = private_key.decrypt(
            cipher_text,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None
            )
        )
        return plain_text.decode()



from django.contrib.gis.geoip2 import GeoIP2
from django.http import HttpRequest



class Location:
    
    
    @staticmethod
    def ip_address(req: HttpRequest) -> str:
        x_forwarded_for = req.META.get('HTTP_X_FORWARDED_FOR')
        if x_forwarded_for:
            ip = x_forwarded_for.split(',')[0]
        else:
            ip = req.META.get('REMOTE_ADDR')
            
        return str(ip)
    
    
    
    @staticmethod
    def geo_data(req: HttpRequest) -> str:
        
        #User Agent
        user_agent = req.META['HTTP_USER_AGENT']

        if 'User-Agent' in req.headers:
            user_agent = req.headers['User-Agent']

        if 'user-agent' in req.headers:
            user_agent = req.headers['user-agent']

        # IP Address
        user_ip = Location.ip_address(req)
        if user_ip.startswith("192.168.") or user_ip.endswith(".0.0.1"):
            user_ip = "154.160.22.132"

        g = GeoIP2()
        city = g.city(user_ip)

        return {
            "user_ip": user_ip,
            "user_agent": user_agent,
            "city": city['city'],
            "continent_code": city['continent_code'],
            "continent_name": city['continent_name'],
            "country_code": city['country_code'],
            "country_name": city['country_name'],
            "is_in_eu": city['is_in_european_union'],
            "latitude": city['latitude'],
            "longitude": city['longitude'],
            "postal_code": city['postal_code'],
            "region": city['region'],
            "time_zone": city['time_zone'],
        }


import requests
import base64
import json
from django.conf import settings
from django.http import HttpRequest

from nkunyim_util.encryption import Encryption
from nkunyim_util.httpsession import HttpSession


class HttpClient:

    def __init__(self, req: HttpRequest, name:str) -> None:
        
        headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/x-www-form-urlencoded'
        }

        try:
            base_url = settings.NKUNYIM_SERVICES[name.upper()]
            sess = HttpSession(req=req)
            
            user_data = sess.get_user()
            if user_data and 'id' in user_data:
                
                plain_text = json.dumps(user_data)
                
                encryption = Encryption()
                cipher_text = encryption.rsa_encrypt(plain_text=plain_text, name=name)
                
                access_token = base64.b64encode(cipher_text)
                headers['Authorization'] = f"JWT {access_token}"
                
        except KeyError as e:
            raise Exception(f"The service configuration variable {name.upper()} has not defined. Error detail: {str(e)}")

        except Exception as ex:
            raise Exception(f"Exception error occured when initializing the HttpClient. Error detail: {str(ex)}")
        
        self.base_url = base_url
        self.headers = headers


    def post(self, path: str, data: dict) -> requests.Response:
        url = self.base_url + path
        return requests.post(url=url, data=data, headers=self.headers)


    def get(self, path: str) -> requests.Response:
        url = self.base_url + path
        return requests.get(url=url, headers=self.headers)


    def put(self, path: str, data: dict) -> requests.Response:
        url = self.base_url + path
        return requests.put(url=url, data=data, headers=self.headers)


    def delete(self, path: str) -> requests.Response:
        url = self.base_url + path
        return requests.delete(url=url, headers=self.headers)
    


from datetime import datetime
from typing import Self, Type
from uuid import UUID

from django.conf import settings
from django.core.paginator import Paginator
from rest_framework.serializers import ModelSerializer

from nkunyim_util.validation import is_uuid

class Pagination:
    
    def __init__(self, serializer: Type[ModelSerializer]):
        
        # ViewSet
        self.serializer = serializer
        
        # Query
        self.rows = 0
        self.page = 0
        self.params = None
        self.path = None
        self.queryset = None
        
        # Here
        self.data = None

    def build(self, key:str, typ:str, val:any) -> None:
        if not self.params:
            self.params = {}
        
        if typ not in ['int', 'bool', 'str', 'uuid', 'float', 'date', 'time', 'timez', 'datetime']:
            self.params[key] = val
            
        if typ == 'uuid' and is_uuid(val=val):
            self.params[key] = UUID(val)
    
        if typ == 'bool':
            self.params[key] = bool(val)
    
        if typ == 'str':
            self.params[key] = str(val)
            
        if typ == 'int':
            self.params[key] = int(val)
            
        if typ == 'float':
            self.params[key] = float(val)
            
        if typ == 'date':
            self.params[key] = datetime.strptime(val, '%Y-%m-%d %H:%M:%S').date()
            
        if typ == 'time':
            self.params[key] = datetime.strptime(val, '%Y-%m-%d %H:%M:%S').time()
            
        if typ == 'timez':
            self.params[key] = datetime.strptime(val, '%Y-%m-%d %H:%M:%S').timestamp()
            
        if typ == 'datetime':
            self.params[key] = datetime.strptime(val, '%Y-%m-%d %H:%M:%S')
    
    
    def extract(self, schema:dict = None, query_params:dict = None) -> None:
        if query_params:
            self.rows = int(query_params.get('rows', 0))
            self.page = int(query_params.get('page', 0))
            
        if self.rows == 0:
            self.rows = settings.REST_FRAMEWORK['PAGE_SIZE'] if settings.REST_FRAMEWORK and 'PAGE_SIZE' in settings.REST_FRAMEWORK else 25
            
        if self.page == 0:
            self.page = 1
            
        if schema:
            for key in list(schema.keys()):
                if query_params and key in query_params:
                    typ = schema[key]
                    self.build(key=key, typ=typ, val=query_params[key])
             

        
    def get(self, count:int = 1, many:bool = False, _prev:str = None, _next:str = None) -> Self:
        
        serialize = self.serializer(self.queryset, many=many)

        if many:
            self.data = {
                'count': count,
                'next': _next,
                'prev': _prev,
                'data': serialize.data
            }
        else:
            self.data = serialize.data
            
        return self
    
    
    def list(self) -> Self:
        paginator = Paginator(self.queryset, int(self.rows))
        self.queryset = paginator.page(int(self.page))
        
        query_params = ""
        if self.params:
            for key in dict(self.params).keys():
                query_params += f"&{key}={self.params[key]}"
                
        _next = f"{settings.APP_BASE_URL}/{self.path}?rows={self.rows}&page={self.page + 1}{query_params}" if self.queryset.has_next() else None
        _prev = f"{settings.APP_BASE_URL}/{self.path}?rows={self.rows}&page={self.page - 1}{query_params}" if self.queryset.has_previous() else None
        
        return self.get(count=paginator.count, many=True, _prev=_prev, _next=_next)
    


from typing import Union
from uuid import uuid4
from django.conf import settings
from django.contrib.messages import get_messages
from django.http import HttpRequest


class HttpSession:

    def __init__(self, req: HttpRequest) -> None:
        self.req = req

    
    def get_host_parts(self) -> list[str]:
        host = self.req.get_host()
        host_parts = host.lower().split('.')
        return host_parts
    
    
    def get_subdomain(self) -> str:
        host_parts = self.get_host_parts()
        return host_parts[-3] if len(host_parts) > 2 else "www"
    
    
    def get_domain(self) -> str:
        host_parts = self.get_host_parts()
        return f"{host_parts[-2]}.{host_parts[-1]}"
    
    
    def get_session_key(self) -> str:
        subdomain = self.get_subdomain()
        domain = self.get_domain()
        session_key = f"{subdomain}.{domain}"
        return session_key
    
        
    def set_account(self, data: any) -> None:
        self.req.session[self.get_session_key()] = data
        self.req.session.modified = True


    def get_account(self) -> Union[dict, None]:
        if not bool(self.get_session_key() in self.req.session):
            return None

        account = self.req.session[self.get_session_key()]
        return account


    def get_user(self) -> Union[dict, None]:
        user_data = None
        if self.req.user.is_authenticated:
            user_data = {
                "id": str(self.req.user.id),
                "username": self.req.user.username,
                "nickname": self.req.user.nickname,
                "first_name": self.req.user.first_name,
                "last_name": self.req.user.last_name,
                "phone_number": self.req.user.phone_number,
                "email_address": self.req.user.email_address,
                "photo_url": self.req.user.photo_url,
                "is_verified": self.req.user.is_verified,
                "is_active": self.req.user.is_active,
                "is_admin": self.req.user.is_admin
            }
            if self.req.user.is_superuser:
                user_data['is_superuser'] = self.req.user.is_superuser
            
        return user_data
    
    
    def get_app_data(self) -> dict:
        account = self.get_account()
        return account['app'] if account and 'app' in account else {
            "name": "Nkunyim",
            "title": "ERP Books",
            "caption": "NKUNYIM Kasa, Amandze, Abakɔsɛm Na Kodzisɛm",
            "description": "Building The Future of Earth's Most Interesting Language",
            "keywords":"",
        }
    
    
    def init_menus(self) -> list[dict]:

        xvix_mid = str(uuid4())
        xvix_menus = [{
            "id": str(uuid4()),
            "node": "system", 
            "module": {
                "id": xvix_mid,
                "name": "Xvix",
                "title": "AutoFix",
                "caption": "Manage auto-fix data",
                "icon": "mdi mdi-auto-fix",
                "path": "xvix",
                "route": "#xvix",
            },
            "items": [],
            "is_active": True
        }]

        return xvix_menus
    

    def make_pages(self, menus: list[dict]) -> tuple:

        toolbox_name, manage_name, system_name = "toolbox", "manage", "system"
        toolbox_menus, manage_menus, system_menus = [], [], []

        pages = dict(settings.NKUNYIM_PAGES)

        env = settings.NKUNYIM_ENV

        for menu in menus:
            # Node
            if menu['node'] == toolbox_name:
                toolbox_menus.append(menu)
                
            if menu['node'] == manage_name:
                manage_menus.append(menu)
                
            if menu['node'] == system_name:
                system_menus.append(menu)

            # Menu
            m_name = menu['module']['name']
            m_path = menu['module']['path']
            m_key = "{}Page".format(str(m_name).title())
            m_val = "./{m}/home.{e}".format(m=str(m_path).lower(), e=env)
            pages[m_key] = m_val

            # Item
            if menu['items'] and len(menu['items']) > 0:
                for item in menu['items']:
                    i_name = item['page']['name']
                    i_path = item['page']['path']
                    i_key = "{}{}Page".format(str(m_name).title(), str(i_name).title())
                    i_val = "./{m}/{p}.{e}".format(m=str(m_path).lower(), p=str(i_path).lower(), e=env)
                    pages[i_key] = i_val
        
        navs = []

        if len(toolbox_menus) > 0:
            navs.append(
                {
                    "name": toolbox_name.title(),
                    "menus": toolbox_menus
                }
            )

        if len(manage_menus) > 0:
            navs.append(
                {
                    "name": manage_name.title(),
                    "menus": manage_menus
                }
            )
            
        if len(system_menus) > 0:
            navs.append(
                {
                    "name": system_name.title(),
                    "menus": system_menus
                }
            )

        return navs, pages
    
       
    
    def get_context(self) -> dict:
        path = self.req.path.lower()
        paths = ['/']
        node = "index"

        if len(path) > 1 and path.strip('/') != "":
            paths = path.strip('/').split('/')
            node = paths[-1]

        # Account
        role_data = None
        business_data = None

        menus_data = []

        user_data = self.get_user()
        if user_data and 'username' in user_data:
            
            account = self.get_account()
            if account and 'role' in account:
                menus_data.extend(account['menus'])
                role_data = account['role']
                business_data = account['business']
                
            # Do it last
            if 'is_superuser' in user_data:
                menus_data.extend(self.init_menus())

        navs_data, pages_data = self.make_pages(menus_data)

        # Messaes
        messages = get_messages(request=self.req)
        alert = {}
        alert_map = {
            "debug": "primary",
            "info": "info",
            "success": "success",
            "warning": "warning",
            "error": "danger",
        }
        for message in messages:
            alert["type"] = alert_map[message.tags]
            alert["message"] = str(message)
            
        context_data = self.get_app_data()
        context_data['static'] = settings.STATIC_URL
        context_data['alert'] = alert
        
        
        if self.get_subdomain() == "app":
            context_data['page'] = {
                "path": path,
                "paths": paths,
                "node": node,
                'name': "{}Page".format(node.title()),
                "colour": "nkunyim"
            }
            context_data['pages'] = pages_data
            context_data['navs'] = navs_data
            context_data['role'] = role_data
            context_data['business'] = business_data
            context_data['user'] = user_data
            
        
        if self.get_subdomain() == "api":
            context_data['page'] = {
                "colour": "nkunyim"
            }
            
        return context_data
 

    def kill(self) -> None:
        self.req.session.clear()
        self.req.session.flush()


from .encryption import Encryption
from .httpclient import HttpClient
from .httpsession import HttpSession
from .location import Location
from .pagination import Pagination
from .messaging import Messaging
from .query import Query
from .validation import Validation